/*
  This file is licensed to You under the Apache License, Version 2.0
  (the "License"); you may not use this file except in compliance with
  the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
*/

package org.xmlunit.diff;

import java.io.File;
import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;

import org.w3c.dom.Document;
import org.xmlunit.diff.Comparison.Detail;

/**
 * The Diff-Object is the result of two comparisons.
 * @see org.xmlunit.builder.DiffBuilder
 */
public class Diff {

    private final Iterable<Difference> differences;

    private final Source controlSource;

    private final Source testSource;

    private static final ComparisonFormatter DEFAULT_FORMATTER =
        new DefaultComparisonFormatter();

    public Diff(Source controlSource, Source testSource, Iterable<Difference> differences) {
        this.controlSource = controlSource;
        this.testSource = testSource;
        this.differences = differences;
    }

    /**
     * @return true if there was at least one difference.
     */
    public boolean hasDifferences() {
        return differences.iterator().hasNext();
    }

    /**
     * @return all differences found before the comparison process stopped.
     */
    public Iterable<Difference> getDifferences() {
        return differences;
    }


    public Source getControlSource() {
        return controlSource;
    }


    public Source getTestSource() {
        return testSource;
    }

    @Override
    public String toString() {
        return toString(DEFAULT_FORMATTER);
    }

    public String toString(ComparisonFormatter formatter) {
        if (!hasDifferences()) {
            return "[identical]";
        }
        return getDifferences().iterator().next().getComparison().toString(formatter);
    }
    
    public static void main( String[] args ) throws Exception {
    	DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
    	dbf.setIgnoringComments( true );
    	dbf.setIgnoringElementContentWhitespace( true );
    	DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
    	File f1 = new File( "d:/travail/soft/japisoft-editix-2015/test/diff/test1.xml" );
    	File f2 = new File( "d:/travail/soft/japisoft-editix-2015/test/diff/test2.xml" );
    	Document doc1 = db.parse( new FileInputStream( f1 ) );
    	Document doc2 = db.parse( new FileInputStream( f2 ) );
    	
    	doc1.normalizeDocument();
    	doc2.normalizeDocument();
    	
    	DOMDifferenceEngine dde = new DOMDifferenceEngine();
    	dde.addDifferenceListener( new ComparisonListener() {
			@Override
			public void comparisonPerformed(Comparison comparison,
					ComparisonResult outcome) {
				Detail left = comparison.getControlDetails();
				System.out.println( "Left:" + left.getXPath() );
				Detail right = comparison.getTestDetails();
				System.out.println( "Right:" + right.getXPath() );
				System.out.println( "--- " + comparison + " => " + outcome );
			}
		}); 
    	
    	dde.compare( new DOMSource( doc1 ), new DOMSource( doc2 ) );
    	
    	
    }

}
