package com.japisoft.dtdparser.node;

import java.io.PrintWriter;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.tree.TreeNode;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class DTDNode implements TreeNode {
	private String value;

	private int type;

	/** Comment type */
	public static final int COMMENT = 0;

	/** Entity type */
	public static final int ENTITY = 1;

	/** Element type */
	public static final int ELEMENT = 2;

	/** Attribute type */
	public static final int ATTRIBUTE = 3;

	/** Root type */
	public static final int ROOT = 4;

	/** Element set */
	public static final int ELEMENT_SET = 5;

	/** Element set content */
	public static final int ELEMENT_REF = 6;

	/** Reset the node type : COMMENT, ENTITY, ELEMENT or ATTRIBUTE */
	public void setNodeType(int type) {
		this.type = type;
	}

	/** @return the node type : COMMENT, ENTITY, ELEMENT or ATTRIBUTE */
	public int getNodeType() {
		return type;
	}
	
	public String toString() {
		return value + "(" + type + ")";
	}

	private RootDTDNode root;

	/** Reset the root node */
	public void setRoot(RootDTDNode node) {
		this.root = node;
	}

	/** @return the root node */
	public RootDTDNode getRoot() {
		return root;
	}

	/** @return true for comment node */
	public boolean isComment() {
		return type == COMMENT;
	}

	/** @return true for entity node */
	public boolean isEntity() {
		return type == ENTITY;
	}

	/** @return true for tag description */
	public boolean isElement() {
		return type == ELEMENT;
	}

	/** @return true for attribute description */
	public boolean isAttribute() {
		return type == ATTRIBUTE;
	}

	/** @return true if this is the root node */
	public boolean isRoot() {
		return type == ROOT;
	}

	/** @return true for a set of element */
	public boolean isElementSet() {
		return type == ELEMENT_SET;
	}

	/** @return true for a part of set of element */
	public boolean isElementRef() {
		return type == ELEMENT_REF;
	}

	private String comment = null;

	public void addNodeComment(String value) {
		if (comment == null) {
			comment = value;
		} else
			comment += "\n" + value;
	}

	public String getNodeComment() {
		return comment;
	}

	public void writeComment(PrintWriter output) {
		if (comment != null) {
			output.write(getDTDComment());
		}
	}

	protected String getDTDComment( String comment ) {
		if (comment == null)
			return "";
		return "<!--" + comment + "-->\n";
	}
	
	protected String getDTDComment() {
		return getDTDComment( comment );
	}

	private Vector children = null;

	/** @return true for empty node */
	public boolean isEmpty() {
		return (children == null);
	}

	/**
	 * Add a DTDNode child /
	 * 
	 * @return the added node
	 */
	public void addDTDNode(DTDNode node) {
		if (children == null)
			children = new Vector();
		children.add(node);
		node.setDTDParentNode(this);
	}

	/** Remove a DTDNode child */
	public void removeDTDNode(DTDNode node) {
		if (children != null)
			children.remove(node);
	}

	/** @return the DTDNode children count */
	public int getDTDNodeCount() {
		if (children == null)
			return 0;
		return children.size();
	}

	/** @return a DTDNode child at index from 0 to DTDNodeCount - 1 */
	public DTDNode getDTDNodeAt(int index) {
		if (children == null)
			return null;
		return (DTDNode) children.get(index);
	}

	/** @return the index of the node parameter or -1 for leaf case */
	public int getDTDNodeIndex(DTDNode node) {
		if (children == null)
			return -1;
		return children.indexOf(node);
	}

	/** @return all children */
	public Enumeration getDTDNodes() {
		if (children == null)
			return null;
		return children.elements();
	}

	private DTDNode parent;

	/** Reset the parent node */
	public void setDTDParentNode(DTDNode parent) {
		this.parent = parent;
	}

	/** @return the parent node */
	public DTDNode getDTDParentNode() {
		return parent;
	}

	/** @return all children matching the type */
	public Enumeration getDTDNodeForType(int type) {
		Vector v = new Vector();
		for (int i = 0; i < getDTDNodeCount(); i++) {
			DTDNode node = getDTDNodeAt(i);
			if (node.getNodeType() == type)
				v.add(node);
		}
		return v.elements();
	}

	// JTREE COMPATIBILITY
	
	public Enumeration children() {
		return getDTDNodes();
	}

	public boolean getAllowsChildren() {
		return false;
	}

	public TreeNode getChildAt(int childIndex) {
		return getDTDNodeAt( childIndex );
	}

	public int getChildCount() {
		return getDTDNodeCount();
	}

	public int getIndex(TreeNode node) {
		return getDTDNodeIndex( (DTDNode)node );
	}

	public TreeNode getParent() {
		return getDTDParentNode();
	}

	public boolean isLeaf() {
		return getChildCount() == 0;
	}
	
	private boolean flagged = false;;
	
	public void setFlagged() { flagged = true; }
	public boolean isFlagged() { return flagged; }
}
