package com.japisoft.dtdparser.node;

import java.io.*;
import java.util.Hashtable;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class ElementSetDTDNode
	extends ElementDTDNode
	implements DocumentWritable {

	public ElementSetDTDNode() {
		super();
		setNodeType(ELEMENT_SET);
	}

	private Hashtable htElementRef;

	/** ( element1 | element2 ) */
	public static int CHOICE_TYPE = 0;
	/** ( element1 , element2 ) */
	public static int SEQUENCE_TYPE = 1;

	private int type;

	/** Set the type of the set :
	 * CHOICE_TYPE or SEQUENCE_TYPE */
	public void setType(int type) {
		this.type = type;
	}

	/** @return the type of the set */
	public int getType() {
		return type;
	}

	/** @return <code>true</code> if the nodeName is supported, previousNode is the previous child, occurence if for the child count 
	<pre>
	&lt;A&gt;
	&lt;B&gt; &lt;/B&gt;
	&lt;C&gt; &lt;/C&gt;
	&lt;/A&gt;
	</pre>
	B is the previous child node of C. B appears only 1
	so we will invoke as sample <code>isNodeChildSupported( "A", "B", 1 )</code> */
	public boolean isNodeChildSupported(
		String nodeName,
		String previousNode,
		int occurrence) {
		if (htElementRef == null)
			return false;

		ElementRefDTDNode ref = (ElementRefDTDNode) htElementRef.get(nodeName);

		boolean res = false;

		if (ref != null) {
			if (ref.getOperator() == ElementDTDNode.ONE_ITEM_OPERATOR) {
				res = (occurrence == 1);
			} else if (
				ref.getOperator() == ElementDTDNode.ZERO_MORE_ITEM_OPERATOR) {
				res = (occurrence == 1) || (occurrence == 0);
			} else if (
				ref.getOperator() == ElementDTDNode.ONE_MORE_ITEM_OPERATOR) {
				res = (occurrence >= 0);
			} else if (
				ref.getOperator() == ElementDTDNode.ZERO_ONE_ITEM_OPERATOR) {
				res = (occurrence > 0);
			}
		} else {

			// Check for other set child
			for (int i = 0; i < getDTDNodeCount(); i++) {
				if (getDTDNodeAt(i).isElementSet()) {
					if (((ElementSetDTDNode) getDTDNodeAt(i))
						.isNodeChildSupported(nodeName, null, occurrence))
						res = true;
				}
			}
		}

		if ((getType() == CHOICE_TYPE) || (previousNode == null))
			return res;
		else {

			// Get the previous ref
			if (ref == null)
				return false; // To correct later

			int i = getDTDNodeIndex(ref);
			if (i == 0)
				return false;
			for (int j = i - 1; j >= 0; j--) {
				if (getDTDNodeAt(j).isElementRef()) {
					ElementRefDTDNode n = (ElementRefDTDNode) getDTDNodeAt(j);
					if (n.getName().equals(previousNode))
						return true;
				}
			}

			return false;
		}
	}

	/** Overriding for elementRef */
	public void addDTDNode(DTDNode node) {
		if (node.isElementRef()) {
			if (htElementRef == null)
				htElementRef = new Hashtable();
			if ( ( ( ElementRefDTDNode )node ).getName() != null )
				htElementRef.put(((ElementRefDTDNode) node).getName(), node);
		}
		super.addDTDNode(node);
	}

	/** Write a minimal valid XML document */
	public void writeDocument(PrintWriter output) throws IOException {
		if ((getOperator() == ElementDTDNode.ZERO_ONE_ITEM_OPERATOR)
			|| (getOperator() == ElementDTDNode.ZERO_MORE_ITEM_OPERATOR))
			return;
		if (getType() == CHOICE_TYPE) {
			// Choose the first one
			for (int i = 0; i < getDTDNodeCount(); i++) {
				if (getDTDNodeAt(i) instanceof DocumentWritable) {
					((DocumentWritable) getDTDNodeAt(i)).writeDocument(output);
					break;
				}
			}

		} else { // Write all
			for (int i = 0; i < getDTDNodeCount(); i++) {
				if (getDTDNodeAt(i) instanceof DocumentWritable) {
					((DocumentWritable) getDTDNodeAt(i)).writeDocument(output);
				}
			}
		}
	}

	public String toString() {
		StringBuffer sb = new StringBuffer("(");
		for (int i = 0; i < getDTDNodeCount(); i++) {
			if (i > 0) {
				if (getType() == CHOICE_TYPE)
					sb.append("|");
				else
					sb.append(",");
			}

			DTDNode node = getDTDNodeAt(i);
			if (!(node instanceof ElementSetDTDNode)) {
				if (node instanceof ElementDTDNode) {
					ElementDTDNode e = (ElementDTDNode) node;
					sb.append(e.getName()).append(e.getOperatorString());
				}
			} else
				sb.append(node.toString());
		}

		sb.append(")").append(getOperatorString());
		return sb.toString();
	}

}

// ElementSetDTDNode ends here
