package com.japisoft.editix.editor.js.kit;

import java.awt.Color;
import java.util.HashMap;

import javax.swing.text.Segment;

import com.japisoft.editix.editor.js.JSSyntaxDescription;
import com.japisoft.framework.preferences.Preferences;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class LineParser {

	private Color valueOfProperty = new Color( 150, 0, 0 );
	private Color defaultColor = Color.BLACK;
	private Color propertyColor = Color.ORANGE.darker();
	private Color idColor = propertyColor;
	private Color commentColor = null;
	private Color stringColor = null;
	private JSSyntaxDescription syntax;
	
	LineParser( String preferenceGroupe, JSSyntaxDescription syntax ) {
		this.syntax = syntax;
		initColorTable( preferenceGroupe );
	}

	private void initColorTable( String preferenceGroupe ) {

		Color c = new Color( 0, 150, 0 );
		c = Preferences.getPreference( preferenceGroupe, "delimitersColor", c );

		for ( int i = 0; i < syntax.getDelimiters().length; i++ )
			mapOfColor.put( Character.toString( syntax.getDelimiters()[ i ] ), c );

		c = new Color( 0, 0, 150 );
		c = Preferences.getPreference( preferenceGroupe, "keywordsColor", c );
		for ( int i = 0; i < syntax.getKeywords().length; i++ ) {
			mapOfColor.put( syntax.getKeywords()[ i ], c );
		}
		commentColor = Preferences.getPreference( preferenceGroupe, "commentColor", Color.DARK_GRAY );		
		mapOfColor.put( "/*", commentColor );
		mapOfColor.put( "//", commentColor );
		defaultColor = Preferences.getPreference( preferenceGroupe, "defaultColor", defaultColor );
		propertyColor = Preferences.getPreference( preferenceGroupe, "propertyColor", propertyColor );
		stringColor = Preferences.getPreference( preferenceGroupe, "stringColor", Color.RED );		
	}
	
	private int tokenCount;
	private Token[] maxTokens = new Token[ 100 ];

	public int getTokenCount() {
		return tokenCount;
	}

	private HashMap mapOfLastTokens = new HashMap();
	private HashMap mapOfColor = new HashMap();
	private StringBuffer sb = new StringBuffer();

	private void resetBuffer() {
		sb.delete( 0, sb.length() + 1 );
	}
	
	public Token[] getTokenForLine(Segment str, int line) {
		tokenCount = 0;

		Token previousLastToken = ( Token )mapOfLastTokens.get(
				new Integer( line - 1 ) 
		);

		resetBuffer();
		boolean commentMode = false;
		char stringMode = 0;
		
		// Restore comment color
		if ( previousLastToken != null ) {
			if ( previousLastToken.color == commentColor ) {
				commentMode = true;	
			} else
			if ( previousLastToken.color == stringColor ) {
				stringMode = previousLastToken.text.charAt( 0 );
			}
		}
		
		
		char oldc = 0;
		boolean lineCommentMode = false;
		char c = 0;

		for ( int i = 0; i < str.length(); i++ ) {

			oldc = c;
			c = str.charAt( i );

			if ( !commentMode && stringMode == 0 ) {
				// Comment opening flag
				if ( syntax.isStartingLinesComment( c, oldc ) ) {
					sb.deleteCharAt( sb.length() - 1 );
					addToken( sb.toString() );
					resetBuffer();
					addToken( "/*" );
					commentMode = true;
					lineCommentMode = false;
					continue;
				} else
				if ( syntax.isStartingLineComment( c, oldc ) ) { 
					sb.deleteCharAt( sb.length() - 1 );
					addToken( sb.toString() );
					resetBuffer();
					addToken( "//" );
					commentMode = true;
					lineCommentMode = true;
					continue;					
				} else
				if ( syntax.isStringDelimiter( c ) ) {
					addToken( sb.toString() );
					resetBuffer();
					sb.append( c );
					stringMode = c;
					continue;
				}
			} else 
			if ( commentMode ) {				
				if ( lineCommentMode ) {
					sb.append( c );
				} else {
					if ( syntax.isStoppingLinesComment( c, oldc ) ) {
						// Comment closing flag
						sb.append( c );
						addToken( sb.toString(), commentColor );
						resetBuffer();
						commentMode = false;
					} else
						sb.append( c );
				}
				continue;
			} else
			if ( stringMode != 0 ) {
				if ( syntax.isStringDelimiter( c ) && ( c == stringMode ) ) {
					sb.append( c );
					addToken( sb.toString(), stringColor );
					resetBuffer();
					
					stringMode = 0;
				} else {
					sb.append( c );
				}
				continue;
			}

			if (  ( c == ' ' ) ||
						( c == '\t' ) ||
							syntax.isDelimiter( c ) ) {
				if ( sb.length() > 0 ) {
					addToken( sb.toString() );
				}
				addToken( "" + c );
				resetBuffer();
				continue;
			}

			sb.append( c );	
		}

		Color col = null;
		if ( commentMode )	// Force a gray
			col = commentColor;
		if ( stringMode != 0 )
			col = stringColor;

		addToken( sb.toString(), col );

		if ( ( commentMode && !lineCommentMode ) || stringMode != 0 ) {
			// For next usage
			mapOfLastTokens.put( 
					line, 
					new Token( "" + stringMode, col ) );			
		} else {
			// Remove the last comment
			if ( mapOfLastTokens.containsKey( line ) ) {
				mapOfLastTokens.remove( line );
			}
		}
		return maxTokens;
	}

	Token addToken( String token ) {
		return addToken( token, null );
	}
	
	Token addToken( String token, Color forceColor ) {
		Color c = forceColor;
		if ( c == null )
			c = ( Color )mapOfColor.get( token );
		
		if ( tokenCount == maxTokens.length ) {
			Token[] tmp = new Token[ maxTokens.length * 2 ];
			System.arraycopy(
					maxTokens,0,tmp,0,maxTokens.length
			);
			maxTokens = tmp;
		}

		Token t = null;

		// Not a keyword
		if ( c == null ) {			
			c = defaultColor;
		}
			
		t = new Token( token, c );
		maxTokens[ tokenCount++ ] = t;
		return t;
	}

	void dump() {
		System.out.println();
		for ( int i = 0; i < tokenCount; i++ ) {
			System.out.print( "[" + maxTokens[ i ] + "]" );
		}
	}

	public static void main( String[] args ) {
		String s = "//a";
		LineParser sl = new LineParser( "JSEditor", new JSSyntaxDescription() );
		String[] tmp = s.split( "\n" );
		for ( int i = 0; i < tmp.length; i++ ) {
			sl.getTokenForLine( new Segment( tmp[ i ].toCharArray(), 0, tmp[ i ].length() ), i );
			sl.dump();
		}
	}
	
}
