package com.japisoft.editix.ui.panels.project2.synchro;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import org.apache.commons.net.ftp.FTPClient;
import org.apache.commons.net.ftp.FTPClientConfig;
import org.apache.commons.net.ftp.FTPFile;

import com.japisoft.editix.ui.panels.project2.Node;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class FTPSynchronizer extends AbstractSynchronizer {
	
	public String getName() {
		return "FTP";
	}

	public void uploadIt( 
			File rootPath, Node node ) throws IOException {
		FTPClient client = connect();
		if ( client == null )
			throw new IOException( "Can't connect" );
		try {
			upload( client, rootPath, node.getPath() );
		} finally {
			client.disconnect();
		}
	}

	private void upload( 
		FTPClient client, 
		File rootPath, 
		File target ) throws IOException {

		String relativePath = getRelativePath( rootPath, target );
		int i = relativePath.lastIndexOf( "/" );
		String name = relativePath;
		if ( i > -1 ) {
			String parent = relativePath.substring( 0, i );
			name = relativePath.substring( i + 1 );
			cd( client, parent );
		}		

		if ( target.isFile() ) {
			FileInputStream input = new FileInputStream( target );
			try {
				info( "Processing " + name );
				client.storeFile( name, input );
			} finally {
				input.close();
			}
		} else {
			String[] files = target.list();
			if ( files != null ) {
				for ( String file : files ) {
					upload( client, rootPath, new File( target, file ) );
				}
			}
		}
	}

	private void cd( FTPClient client, String path ) throws IOException {
		client.changeWorkingDirectory( getFTPRootPath() );
		String[] parts = path.split( "/" );
		for ( String part : parts ) {
			if ( "".equals( part ) )
				continue;
			client.makeDirectory( part );
			client.changeWorkingDirectory( part );
		}
	}
		
	public void downloadIt( 
			File rootPath, Node node ) throws IOException {
		FTPClient client = connect();
		if ( client == null )
			throw new IOException( "Can't connect" );
		try {
			download( client, rootPath, node.getPath() );
		} finally {
			client.disconnect();
		}				
	}

	private void download( 
		FTPClient client, File rootPath, File target ) throws IOException {

		String relativePath = getRelativePath( rootPath, target );
		String name = relativePath;
		
		if ( target.isDirectory() ) {
			cd( client, relativePath );
			name = null;
		} else {
			int i = relativePath.lastIndexOf( "/" );
			if ( i > -1 ) {
				String parent = relativePath.substring( 0, i );
				name = relativePath.substring( i + 1 );
				cd( client, parent );
			}
		}
		
		FTPFile[] files = client.listFiles( name );
		if ( files != null ) {
			for ( FTPFile file : files ) {
				if ( ".".equals( file.getName() ) )
					continue;
				if ( "..".equals( file.getName() ) )
					continue;				
				info( "Processing " + file.getName() );
				if ( file.isFile() ) {
					File f = target;
					if ( f.isDirectory() )
						f = new File( f, file.getName() );
					FileOutputStream output = new FileOutputStream( f );
					try {
						client.retrieveFile( file.getName(), output );
					} finally {
						output.close();
					}
				} 
			}
		}
		files = client.listDirectories( name );
		if ( files != null ) {
			for ( FTPFile file : files ) {
				if ( ".".equals( file.getName() ) )
					continue;
				if ( "..".equals( file.getName() ) )
					continue;
				target = new File( target, file.getName() );
				target.mkdir();
				download( client, rootPath, target );
			}
		}
	}		

	public boolean test() {
		FTPClient client = null;
		try {
			client = connect();
			return client != null;
		} catch( IOException exc ) {
			if ( client != null ) {
				try {
					client.disconnect();
				} catch( Exception exc2 ) {}
			}
		}
		return false;
	}

	private FTPClient connect() throws IOException {
		FTPClient client = new FTPClient();
		FTPClientConfig config = new FTPClientConfig();
		client.configure( config );
		try {
			client.connect( 
				getProperty( "host" ), 
				Integer.parseInt( getProperty( "port" ) ) 
			);
			client.login( getProperty( "user" ), getProperty( "password" ) );
			client.setFileType( FTPClient.BINARY_FILE_TYPE );
			client.enterLocalPassiveMode();
			if ( !client.changeWorkingDirectory( getFTPRootPath() ) )
				return null;
			return client;
		} catch( Exception exc ) {
			client.disconnect();
			return null;
		}
	}

	private String getFTPRootPath() {
		String path = getProperty( "path" );
		if ( path == null || "".equals( path ) )
			path = "/";
		return path;
	}
	
	@Override
	public String toString() {
		return super.toString() + " [" + getProperty( "host" ) + "]";
	}
	
	public static void main( String[] args ) {
		FTPSynchronizer t = new FTPSynchronizer();
		t.setProperty( "host", "www.unimailer.com" );
		t.setProperty( "port", "21" );
		t.setProperty( "user", "test@unimailer.com" );
		t.setProperty( "password", "abrillant" );
		t.setProperty( "path", "/aa" );
		System.out.println( "TEST=" + t.test() );
	}

}
