package com.japisoft.framework;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import com.japisoft.framework.log.Logger;
import com.japisoft.framework.preferences.Preferences;
import com.japisoft.framework.step.Threadable;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class ApplicationMain {


	private static String[] initArgs;
	private static List<ApplicationStepListener> listeners = null;
	
	public static void addApplicationStepListener( ApplicationStepListener listener ) {
		if ( listeners == null )
			listeners = new ArrayList<ApplicationStepListener>();
		listeners.add( listener );
	}

	public static void removeApplicationStepListener( ApplicationStepListener listener ) {
		if ( listeners != null )
			listeners.remove( listener );
	}
	
	public static void fireApplicationStepEvent( ApplicationStep step, int indice, int maxIndice ) {
		if ( listeners != null ) {
			for ( ApplicationStepListener listener : listeners ) {
				listener.run( step, indice, maxIndice );
			}
		}
	}

	/** Start the application using the following application descriptor file 
	 * @param applicationDescriptorFile A path using the CLASSPATH for the application descriptor file
	 * @param args The application main parameters
	 * @throws ApplicationException when cannot find the application descriptor file */
	public static void start( String applicationDescriptorFile, String[] args ) throws ApplicationException {
		ApplicationModel.init( applicationDescriptorFile );
		start( args );
	}

	/** Start the application without application descriptor file. The user must
	 * update each correct field from the application model before calling it
	 * @param args The application main parameters */
	public static void start( String[] args ) throws ApplicationException {
		initArgs = args;
		if ( ApplicationModel.getApplicationStepCount() == 0 ) {
			throw new ApplicationException( "No application step found" );
		}
		boolean started = false;
		try {
			for ( int i = 0; i < ApplicationModel.getApplicationStepCount(); i++ ) {
				ApplicationStep step = ApplicationModel.getApplicationStepAt( i );
				if ( !step.isFinal() ) {
					ApplicationModel.debug(
							"Start " + ( i + "/" + ApplicationModel.getApplicationStepCount() ) + " " + step.getClass() );
					started = true;
					try {
						fireApplicationStepEvent(step, i, ApplicationModel.getApplicationStepCount() );
						startStep( step, args );
					} catch( Exception e ) {
						if ( e instanceof ApplicationStepException ) {
							if ( ( ( ApplicationStepException )e ).isCritical() ) {
								throw new ApplicationException( e );
							}
						} else
							throw new ApplicationException( e );
					}
				}
			}
		} finally {
			// Stop all the non final application step
			for ( int i = 0; i < ApplicationModel.getApplicationStepCount(); i++ ) {
				ApplicationStep step = ApplicationModel.getApplicationStepAt( i );
				if ( !step.isFinal() && !(step instanceof Threadable ) ) {
					try {
						ApplicationModel.debug(
								"Stop " + step.getClass() );						
						step.stop();
					} catch( Throwable th ) {
						Logger.addException( th );
					}
					
				}
			}
			
			listeners = null;
			
		}
		if ( !started )
			throw new ApplicationException( "No application step found. Only final application step available ?" );
	}

	private static void startStep( ApplicationStep step, String[] args ) throws Exception {
		if ( step instanceof Threadable ) {
			// Asynchronous step
			new Thread( new Runnable() {
				
				@Override
				public void run() {
					try {
						step.start( args );
						step.stop();
					} catch( Exception exc ) {
						ApplicationModel.debug( exc );
					}
				}
			} ).start();

		} else {
			step.start( args );
		}
	}
	
	/** Should be called when terminating the application. This method will call the exit method */
	public static void quit( int exitCode ) {
		/* Called any final step */
		for ( int i = 0; i < ApplicationModel.getApplicationStepCount(); i++ ) {
			ApplicationStep step = ApplicationModel.getApplicationStepAt( i );
			if ( step.isFinal() ) {
				try {
					ApplicationModel.debug( "Start and Stop " + step.getClass() );					
					step.start( initArgs );
					step.stop();
				} catch( Throwable th ) {
					Logger.addException( th );
				}
			} else
				step.quit();
		}
		Preferences.savePreferences();
		System.exit( exitCode );
	}
	
	/**
	 * Show a dialog box with the following error
	 * @param th An exception while starting the application */
	public static void showException( Throwable th ) {
		th.printStackTrace();
		StringWriter sw = new StringWriter();
		th.printStackTrace(
				new PrintWriter( sw ) );

		JFrame fr = new JFrame( "Unknown error !" );
		StringBuffer sb = new StringBuffer();
		sb.append( ApplicationModel.LONG_APPNAME + " " + ApplicationModel.getAppVersion()
				+ " - Build " + ApplicationModel.BUILD);
		sb.append("\nHas met an unknown error, please send it to : "
				+ ApplicationModel.MAIN_SUPPORT_EMAIL + "\n");
		sb
				.append("specifying your operating system version and java version\n");
		
		sb.append( "Java Version : ").append( System.getProperty( "java.version" ) ).append( "\n" );
		sb.append( "Operating System : " ).append( System.getProperty( "os.name" ) ).append( "\n" );
		
		sb
				.append("\n----------------------------------------------------------------\n\n");
		sb.append(sw.toString());
		JTextArea ar = new JTextArea( sb.toString() );
		fr.getContentPane().add( new JScrollPane( ar ) );
		fr.pack();
		fr.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		fr.setVisible( true );
	}
	
}
