package com.japisoft.framework.dialog.actions;

import java.util.ArrayList;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class DialogActionModel {

	private ArrayList list;
	private DialogActionModelListener listener;

	public static DialogAction DEFAULT_OKACTION = new OKAction();
	public static DialogAction DEFAULT_CANCELACTION = new CancelAction();

	/** This is a 'OK', 'Cancel' dialog model */
	public static DialogActionModel getDefaultDialogActionModel() {
		return new DialogActionModel(
				new DialogAction[] {
						DEFAULT_CANCELACTION,
						DEFAULT_OKACTION,
				} );
	}

	/** This is an 'OK' dialog model */
	public static DialogActionModel getDefaultDialogOkActionModel() {
		return new DialogActionModel(
				new DialogAction[] {
						DEFAULT_OKACTION,
				} );
	}

	/** This is an 'Close' dialog model */
	public static DialogActionModel getDefaultDialogCloseActionModel() {
		return new DialogActionModel(
				new DialogAction[] {
						new ClosingAction(),
				} );
	}
		
	/** Empty model */
	public DialogActionModel() {}

	/** Unique action model */
	public DialogActionModel( DialogAction action ) {
		addDialogAction( action );
	}
	
	/** Multiple actions model */
	public DialogActionModel( DialogAction[] actions ) {
		if ( actions != null )
			for ( int i = 0; i < actions.length; i++ ) {
				addDialogAction( actions[ i ] );
			}
	}

	public Object clone() {
		DialogActionModel model = new DialogActionModel();
		model.list = new ArrayList();
		if ( list != null )
			for ( int i = 0; i < list.size(); i++ )
				model.list.add( list.get( i ) );
		return model;
	}
	
	/** Listener for each change inside this model */
	public void setModelListener( DialogActionModelListener listener ) {
		this.listener = listener;
	}

	/** @return a dialog action for this id */
	public DialogAction getDialogActionForId( int id ) {
		for ( int i = 0; i < getDialogActionCount(); i++ ) {
			DialogAction action = getDialogActionAt( i );
			if ( action.getActionId() == id )
				return action;
		}
		return null;
	}
	
	/** Check for the following action */
	public boolean hasDialogActionForId( int id ) {
		return ( getDialogActionForId( id ) != null );
	}

	/** 
	 * Add a new dialog action */
	public DialogActionModel addDialogAction( DialogAction action ) {
		if ( list == null )
			list = new ArrayList();
		if ( list.contains( action ) )
			throw new RuntimeException( "Error : You have added several times the same action object" );
		checkForUniqueAction( action );
		list.add( action );
		fireModelUpdated();
		return this;
	}

	private void checkForUniqueAction( DialogAction action ) {
		if ( list != null ) {
			for ( int i = 0;i < list.size(); i++ ) {
				if ( 
						( ( DialogAction )list.get( i ) ).getActionId() ==
								action.getActionId() ) {
					throw new RuntimeException( "Error : Found two actions with the same id " + action.getActionId() );
				}
			}
		}
	}
	
	/**
	 * Add a new dialog action at this index
	 * @param index Value starting from 0
	 * @param action A swing action
	 */
	public void addDialogAction( int index, DialogAction action ) {
		if ( list == null )
			list = new ArrayList();
		checkForUniqueAction( action );		
		list.add( index, action );
		fireModelUpdated();
	}

	/**
	 * Remove this action
	 */
	public void removeDialogAction( DialogAction action ) {
		if ( list != null ) {
			list.remove( action );
			fireModelUpdated();
		}
	}
		
	private void fireModelUpdated() {
		if ( listener != null )
			listener.modelUpdated();
	}
	
	/**
	 * @return The number of dialog action
	 */
	public int getDialogActionCount() { 
		if ( list == null )
			return 0;
		return list.size();
	}
	
	/**
	 * @param index value starting from 0
	 * @return a dialog action at this index
	 */
	public DialogAction getDialogActionAt( int index ) {
		if ( list == null )
			return null;
		return ( DialogAction )list.get( index );
	}

}
