package com.japisoft.framework.xml;

import java.io.File;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import com.japisoft.framework.ApplicationModel;
import com.japisoft.framework.preferences.Preferences;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public final class XMLParser {

	public static final String XMLPARSER_PREFERENCE_NAME = "xmlparser";
	static final String JAXP_SCHEMA_LANGUAGE = "http://java.sun.com/xml/jaxp/properties/schemaLanguage";
	static final String W3C_XML_SCHEMA = "http://www.w3.org/2001/XMLSchema";
	public static String DEF_PARSER_NAME = "XERCES";
	public static String DEF_PARSER_CLASS = "org.apache.xerces.jaxp.SAXParserFactoryImpl";
	public static String DEF_LIBRARY_PATH = "parser";
	private static Class DEF_PARSER = null;
	private static boolean ERROR_FOUND = false;
	private static boolean NAMESPACEAWARE = false;
	private static boolean SCHEMAAWARE = true;
	public static final String CONFIG_FILE = "jaxp_parser.dat";

	// -------------------------------------------------------------------------------------

	public static void resetDefaultParser() {

		NAMESPACEAWARE = Preferences.getPreference("xmlconfig",
				"namespaceAware", true);

		// Case for XERCES
		if (Preferences.getPreference("xmlconfig", "xinclude", true)) {
			System.setProperty(
					"org.apache.xerces.xni.parser.XMLParserConfiguration",
					"org.apache.xerces.parsers.XIncludeParserConfiguration");
		}

		ApplicationModel.debug( "Reset the default parser" );
		
		if ( DEF_PARSER == null && 
				!ERROR_FOUND ) {

			if ( !loadByConfigFile() ) {

				String parser = Preferences.getPreference( 
						Preferences.SYSTEM_GP,
						XMLPARSER_PREFERENCE_NAME, 
						DEF_PARSER_NAME );

				ApplicationModel.debug( "Reset from the preference " + parser );
				
				if ( !DEF_PARSER_NAME.equals( parser ) ) {

					ApplicationModel.debug( "Load from " + DEF_LIBRARY_PATH );
					
					// Search for a library inside the ext/parser directory

					ArrayList al = XMLToolkit.browseJars(DEF_LIBRARY_PATH);

					if (al.size() == 0) {
						try {
							DEF_PARSER = Class.forName(parser);
						} catch (Throwable th) {
							System.err.println("Can't use the " + parser
									+ " class");
							ERROR_FOUND = true;
						}
					} else {

						URL[] libs = new URL[al.size()];
						for (int i = 0; i < al.size(); i++)
							libs[i] = ( URL ) al.get(i);

						try {

							URLClassLoader loader = new URLClassLoader(libs);
							DEF_PARSER = loader.loadClass(parser);
					
							ApplicationModel.debug( "Find " + DEF_PARSER );
							
						} catch (Throwable th) {
							System.err.println("Can't use " + parser + " : "
									+ th.getMessage());
							ERROR_FOUND = true;
						}

					}

				}
			}

		}

	}

	static boolean loadByConfigFile() {
		try {
			ApplicationModel.debug( "Load " + CONFIG_FILE );
			Class cl = XMLToolkit.loadByConfigFile(new File(ApplicationModel
					.getAppUserPath(), CONFIG_FILE));
			if (cl != null) {
				ApplicationModel.debug( "Class found " + cl );
				DEF_PARSER = cl;
				return true;
			} else
				ApplicationModel.debug( "No class found" );
		} catch (Exception exc) {
			ApplicationModel.debug( exc );
		}
		return false;
	}

	public static SAXParserFactory getSAXParserFactory() {

		resetDefaultParser();

		SAXParserFactory factory = null;

		if ( DEF_PARSER != null ) {
			try {
				factory = (javax.xml.parsers.SAXParserFactory) DEF_PARSER
						.newInstance();
			} catch (Throwable th) {
				System.err.println( "Can't use " + DEF_PARSER + " : "
						+ th.getMessage() );
			}
		}

		if ( DEF_PARSER == null ) {
			try {
				DEF_PARSER = Class.forName(DEF_PARSER_CLASS);
				factory = (javax.xml.parsers.SAXParserFactory) DEF_PARSER
						.newInstance();
			} catch (Throwable th) {
				th.printStackTrace();
				throw new RuntimeException("ERROR CAN'T FIND "
						+ DEF_PARSER_CLASS + "???");
			}
		}

		factory.setNamespaceAware( NAMESPACEAWARE );

		ApplicationModel.debug( "Use the factory " + factory );
		
		return factory;
	}

	private static SAXParserFactory FACTORY = null;

	public static SAXParserFactory getSaxParserFactory(boolean validating) throws Exception {
		if ( FACTORY == null )
			FACTORY = getSAXParserFactory();
		
		FACTORY.setValidating( validating );
		try {
			FACTORY.setFeature( 
				"http://apache.org/xml/features/dom/include-ignorable-whitespace",
				false
			);
		} catch( Throwable th ) {}
		return FACTORY;
	}
	
	public static SAXParser getSaxParser(boolean validating) throws Exception {
		
		SAXParser parser = getSaxParserFactory(validating).newSAXParser();

/*		
 * Provoque un BUG dans le cas d'une DTD !!
 * if ( validating )
		try {
			parser.setProperty( JAXP_SCHEMA_LANGUAGE, W3C_XML_SCHEMA );
		} catch (Exception exc) {
		}*/

		ApplicationModel.debug( "Use the parser " + parser + " / validating : " + validating );
		
		return parser;
	}

}
