package com.japisoft.framework.xml.parser.dom;

import org.w3c.dom.*;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class CharacterDataImpl extends NodeImpl implements CharacterData {
    public CharacterDataImpl() {
	super();   
    }

    public CharacterDataImpl( int type ) {
	super( type, null );
    }

    public CharacterDataImpl( int type, String content ) {
	super( type, content );
    }

    /**
     * The character data of the node that implements this interface. The DOM 
     * implementation may not put arbitrary limits on the amount of data that 
     * may be stored in a  <code>CharacterData</code> node. However, 
     * implementation limits may  mean that the entirety of a node's data may 
     * not fit into a single <code>DOMString</code>. In such cases, the user 
     * may call <code>substringData</code> to retrieve the data in 
     * appropriately sized pieces.
     * @exception DOMException
     *   NO_MODIFICATION_ALLOWED_ERR: Raised when the node is readonly.
     * @exception DOMException
     *   DOMSTRING_SIZE_ERR: Raised when it would return more characters than 
     *   fit in a <code>DOMString</code> variable on the implementation 
     *   platform.
     */
    public String getData() throws DOMException {
	return getContent();
    }

    public void setData(String data) throws DOMException {
	setContent( data );
    }

    /**
     * The number of characters that are available through <code>data</code> and 
     * the <code>substringData</code> method below.  This may have the value 
     * zero,  i.e., <code>CharacterData</code> nodes may be empty.
     */
    public int getLength() {
	if ( getContent() == null )
	    return 0;
	return getContent().length();
    }

    /**
     * Extracts a range of data from the node.
     * @param offset Start offset of substring to extract.
     * @param count The number of characters to extract.
     * @return The specified substring. If the sum of <code>offset</code> and 
     *   <code>count</code> exceeds the <code>length</code>, then all 
     *   characters to the end of the data are returned.
     * @exception DOMException
     *   INDEX_SIZE_ERR: Raised if the specified offset is negative or greater 
     *   than the number of characters in <code>data</code>, or if the 
     *   specified <code>count</code> is negative.
     *   <br>DOMSTRING_SIZE_ERR: Raised if the specified range of text does not 
     *   fit into a <code>DOMString</code>.
     */
    public String substringData(int offset, 
				int count)
	throws DOMException {
	if ( getContent() == null )
	    throw new DOMExceptionImpl( DOMException.INDEX_SIZE_ERR, " no content" );
	if ( offset < 0 || ( offset + count ) >= getContent().length() ) {
	    throw new DOMExceptionImpl( DOMException.INDEX_SIZE_ERR, " no content" );
	}
	return getContent().substring( offset, offset + count );
    }

    /**
     * Append the string to the end of the character data of the node. Upon 
     * success, <code>data</code> provides access to the concatenation of 
     * <code>data</code> and the <code>DOMString</code> specified.
     * @param arg The <code>DOMString</code> to append.
     * @exception DOMException
     *   NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     */
    public void appendData(String arg)
	throws DOMException {
	if ( getContent() == null )
	    setContent( arg );
	else
	    setContent( getContent() + arg );
    }

    /**
     * Insert a string at the specified character offset.
     * @param offset The character offset at which to insert.
     * @param arg The <code>DOMString</code> to insert.
     * @exception DOMException
     *   INDEX_SIZE_ERR: Raised if the specified offset is negative or greater 
     *   than the number of characters in <code>data</code>.
     *   <br>NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     */
    public void insertData(int offset, 
			   String arg)
	throws DOMException {
	if ( getContent() == null )
	    throw new DOMExceptionImpl( DOMException.INDEX_SIZE_ERR, "Invalid content" );
	if ( offset >= getContent().length() )
	    throw new DOMExceptionImpl( DOMException.INDEX_SIZE_ERR, "Invalid offset " + offset );
	String s = getContent();
	setContent( s.substring( 0, offset ) + arg + s.substring( offset ) );
    }

    /**
     * Remove a range of characters from the node. Upon success, 
     * <code>data</code> and <code>length</code> reflect the change.
     * @param offset The offset from which to remove characters.
     * @param count The number of characters to delete. If the sum of 
     *   <code>offset</code> and <code>count</code> exceeds <code>length</code> 
     *   then all characters from <code>offset</code> to the end of the data 
     *   are deleted.
     * @exception DOMException
     *   INDEX_SIZE_ERR: Raised if the specified offset is negative or greater 
     *   than the number of characters in <code>data</code>, or if the 
     *   specified <code>count</code> is negative.
     *   <br>NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     */
    public void deleteData(int offset, 
			   int count)
	throws DOMException {
	if ( getContent() == null )
	    throw new DOMExceptionImpl( DOMException.INDEX_SIZE_ERR, "No content" );
	if ( offset < 0 || ( offset + count ) >= getContent().length() )
	    throw new DOMExceptionImpl( DOMException.INDEX_SIZE_ERR, "Invalid offset " + offset );
	String s = getContent();
	setContent( s.substring( 0, offset ) + s.substring( offset + count ) );
    }

    /**
     * Replace the characters starting at the specified character offset with 
     * the specified string.
     * @param offset The offset from which to start replacing.
     * @param count The number of characters to replace. If the sum of 
     *   <code>offset</code> and <code>count</code> exceeds <code>length</code>
     *   , then all characters to the end of the data are replaced (i.e., the 
     *   effect is the same as a <code>remove</code> method call with the same 
     *   range, followed by an <code>append</code> method invocation).
     * @param arg The <code>DOMString</code> with which the range must be 
     *   replaced.
     * @exception DOMException
     *   INDEX_SIZE_ERR: Raised if the specified offset is negative or greater 
     *   than the number of characters in <code>data</code>, or if the 
     *   specified <code>count</code> is negative.
     *   <br>NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
     */
    public void replaceData(int offset, 
			    int count, 
			    String arg)
	throws DOMException {
	if ( getContent() == null )
	    throw new DOMExceptionImpl( DOMException.INDEX_SIZE_ERR, "No content" );
	if ( offset < 0 || ( offset + count ) > getContent().length() ) {
	    throw new DOMExceptionImpl( DOMException.INDEX_SIZE_ERR, "Invalid offset " + offset );
	}
	String s = getContent();
	setContent( s.substring( 0, offset ) + arg + s.substring( offset + count ) );
    }

}

// CharacterDataImpl ends here

