package com.japisoft.framework.xml.xsd.instance;

import java.io.StringReader;
import java.util.List;

import com.japisoft.framework.collection.FastVector;
import com.japisoft.framework.xml.XMLFileData;
import com.japisoft.framework.xml.XMLToolkit;
import com.japisoft.framework.xml.parser.FPParser;
import com.japisoft.framework.xml.parser.node.FPNode;

/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class XSDInstanceGenerator {

	public static String generateXMLInstance( String rootElement, String schemaURI ) throws Throwable {
		
		XMLFileData xfd = 
			XMLToolkit.getContentFromURI( schemaURI, null );
		String fullText = xfd.getContent();
		FPParser p = new FPParser();
		p.setFlatView( true );
		FPNode root = ( FPNode )p.parse(new StringReader(fullText)).getRoot();
		
		String targetNamespace = root.getAttribute( "targetNamespace" );
		if ( targetNamespace != null ) {
			for ( int i = 0; i < root.childCount(); i++ ) {
				root.childAt( i ).setApplicationObject( targetNamespace );
			}
		}

		resolveNamespaceReference( root );
		resolveIncludeRedefineImport( root, schemaURI );
		StringBuffer res = new StringBuffer();
		
		XSDBuildInstance xbi = new XSDBuildInstance();
		xbi.buildElement( res, rootElement, root, schemaURI, true );

		return res.toString();
	}

	private static String resetReferenceName( String name, FPNode root ) {
		int i = name.lastIndexOf( ":" );
		if ( i > -1 ) {
			String prefix = name.substring( 0, i );
			String uri = root.getNameSpaceDeclarationURI( prefix );
			if ( uri != null )
				return "{" + uri + "}" + name.substring( i + 1 );
		}
		return null;
	}
	
	private static void resolveNamespaceReference( FPNode root ) {

		// Replace all the reference with the real namespace

		List<FPNode> fv = root.getDocument().getFlatNodes();
		for ( int j = 0; j < fv.size(); j++ ) {
			
			FPNode sn = ( FPNode )fv.get( j );
			if ( sn.hasAttribute( "base" ) ) {
				
				String newBase = resetReferenceName( sn.getAttribute( "base" ), root );
				if ( newBase != null )
					sn.setAttribute( "base", newBase );

			} else
			if ( sn.hasAttribute( "ref" ) ) {

				String newRef = resetReferenceName( sn.getAttribute( "ref" ), root );
				if ( newRef != null )
					sn.setAttribute( "ref", newRef );
								
			} else
			if ( sn.hasAttribute( "type" ) ) {

				String newType = resetReferenceName( sn.getAttribute( "type" ), root );
				if ( newType != null )
					sn.setAttribute( "type", newType );
								
			}

		}		
		
	}
	
	private static void resolveIncludeRedefineImport( FPNode root, String schemaURI ) throws Throwable {

		String targetNamespace = root.getAttribute( "targetNamespace" );
		
		for ( int i = 0; i < root.childCount(); i++ ) {
			
			if ( root.childAt( i ).matchContent( "include" ) ) {
				resolveInclude( targetNamespace, root.childAt( i ), schemaURI );
			} else
			if ( root.childAt( i ).matchContent( "redefine" ) ) {
				resolveRedefine( targetNamespace, root.childAt( i ), schemaURI );
			} else
			if ( root.childAt( i ).matchContent( "import" ) ) {
				resolveImport( root.childAt( i ), schemaURI );
			}

		}

	}

	private static FPNode resolveSchemaLocation( FPNode node, String schemaURI ) throws Throwable {
		String location = node.getAttribute( "schemaLocation" );
		if ( location != null ) {
			XMLFileData xfd = XMLToolkit.getContentFromRelativeOrAbsoluteLocation( location, schemaURI );
			FPParser p = new FPParser();
			p.setFlatView( true );
			FPNode newRoot = ( FPNode )p.parse(new StringReader(xfd.getContent())).getRoot();
			resolveNamespaceReference( newRoot );
			resolveIncludeRedefineImport( newRoot, xfd.uri );
			return newRoot;
		} else 
			return null;
	}

	private static void resolveInclude( String targetNamespace, FPNode node, String schemaURI ) throws Throwable {
		// Include the root content
		FPNode newRoot = resolveSchemaLocation( node, schemaURI );
		if ( newRoot != null ) {
			FPNode parentNode = node.getFPParent();
			for ( int i = 0; i < newRoot.childCount(); i++ ) {
				newRoot.childAt( i ).setApplicationObject( targetNamespace );
				parentNode.appendChild( newRoot.childAt( i ) );
				
			}
		}
	}

	private static void resolveRedefine( String targetNamespace, FPNode node, String schemaURI ) throws Throwable {
		// Redefine the root content
		FPNode newRoot = resolveSchemaLocation( node, schemaURI );
		if ( newRoot != null ) {
			FPNode parentNode = node.getFPParent();
			for ( int i = 0; i < newRoot.childCount(); i++ ) {
				newRoot.childAt( i ).setApplicationObject( targetNamespace );
				parentNode.appendChild( newRoot.childAt( i ) );
			}
			// Add overriding elements in priority
			for ( int i = 0; i < node.childCount(); i++ ) {
				node.childAt( i ).setApplicationObject( "override" );
				parentNode.insertChildNode( 0, node.childAt( i ) );
			}
		}
	}

	private static void resolveImport( FPNode node, String schemaURI ) throws Throwable {
		// Redefine the root content
		FPNode newRoot = resolveSchemaLocation( node, schemaURI );
		if ( newRoot != null ) {
			FPNode parentNode = node.getFPParent();
			for ( int i = 0; i < newRoot.childCount(); i++ ) {
				// Store the targetNamespace for global definition
				newRoot.childAt( i ).setApplicationObject( newRoot.getAttribute( "targetNamespace" ) );
				parentNode.appendChild( newRoot.childAt( i ) );
			}
		}		
	}
	
}
