package com.japisoft.xmlpad.error;

import java.util.ArrayList;

import com.japisoft.xmlpad.Debug;
/**
EditiX Open source XML Editor
https://www.editix.com


Sources Edition :

Our software is available under both a commercial and a non-commercial license :

- If you are a non-commercial user, you are free to use the software for any non-commercial purpose under the terms of the Creative Commons Attribution-NonCommercial-NoDerivatives 4.0 International Public License. 

- If you are a commercial user and you want to use the software for commercial purposes, you must purchase a commercial license. The details are inside the license-commercial.txt file.



Binaries Edition :

- You can purchase a license for binaries usage using the purchasing page at https://www.editix.com. Each binary version can be downloaded on https://www.editix.com.



Please contact us at https://www.editix.com for more information on purchasing a commercial license.


@author (c) 2010-2023 Alexandre Brillant



*/
public class ErrorManager implements ErrorListener {
	
	public static final String ON_THE_FLY_PARSING_CONTEXT = "onfly";
	
	private ArrayList errors = null;

	/** Add a new listener for onTheFly or full parsing error */
	public void addErrorListener( ErrorListener listener ) {
		if ( errors == null )
			errors = new ArrayList();
		if ( listener != null && !errors.contains( listener ) )
			errors.add( listener );
	}

	public void removeErrorListener( ErrorListener listener ) {
		if ( errors != null )
			errors.remove( listener );
	}

	/** Inform all the listeners, that some error reporting is beginning */
	public void initErrorProcessing() {
		flushLastError();
		Debug.debug( "** Init error processing " );
		if ( errors != null ) {
			for ( int i = 0; i < errors.size(); i++ ) {
				( ( ErrorListener )errors.get( i ) ).initErrorProcessing();
			}
		}
	}

	/** Inform all the listeners, that no more error reporting will be done */
	public void stopErrorProcessing() {
		Debug.debug( "** Stop error processing " );
		if ( errors != null ) {
			for ( int i = 0; i < errors.size(); i++ ) {
				( ( ErrorListener )errors.get( i ) ).stopErrorProcessing();
			}
		}		
	}

	boolean lastCaseInError = false;
	boolean lastCaseOnTheFly = false;
	
	/** Here a shortcut for sending a single error message */
	public void notifyError(
			String message ) {
		notifyError(
				message, 
				0
		);
	}

	/** Here a shortcut for sending a single error message */
	public void notifyError(
			String message, int line, boolean onTheFly ) {
		notifyError(
			null,
			true,
			null,
			line,
			0,
			0,
			message
			,
			onTheFly 
		);		
	}
		
	/** Here a shortcut for sending a single error message at the following line */	
	public void notifyError(
			String message,
			int line ) {
		notifyError(
				null,
				true,
				null,
				line,
				0,
				0,
				message
				,
				true );		
	}

	private String resolverSourceLocationSpace( String location ) {
		if ( location == null )
			return null;
		location = location.replaceAll( "%20", " " );
		return location;
	}
	
	/** Inform all the listeners. It is possible to have multiple errors.
	 * So the errors reporting must be completed by calling the
	 * <code>stopErrorProcessing</code> method.
	 * */	
	public void notifyError(
			Object context,
			boolean localError,
			String sourceLocation, 
			int line,
			int column,
			int offset,
			String message, 
			boolean onTheFly ) {
		
		Debug.debug( 
				"** XML Error : " + 
				message + 
				" local:" + 
				localError + 
				" source:" + sourceLocation +
				" line:" + line +
				" col:" + column +
				" offset:" + offset +
				" onFly:" + onTheFly );
		
		sourceLocation = resolverSourceLocationSpace( sourceLocation );
		
		lastCaseInError = true;
		lastCaseOnTheFly = onTheFly;
		if ( errors != null ) {
			for ( int i = 0; i < errors.size(); i++ ) {
				( ( ErrorListener )errors.get( i ) ).notifyError(
						context,
						localError,
						sourceLocation,
						line,
						column,
						offset,
						message,
						onTheFly );
			}			
		}		
	}

	/** 
	 * Notify only one error. It will called the stopErrorProcessing method
	 * at the end
	 * @param localError
	 * @param sourceLocation
	 * @param line
	 * @param column
	 * @param offset
	 * @param message
	 * @param onTheFly*/
	public void notifyUniqueError(
			boolean localError,
			String sourceLocation, 
			int line,
			int column,
			int offset,
			String message, 
			boolean onTheFly ) {
		notifyError( 
				null,
				localError,
				sourceLocation,
				line,
				column,
				offset,
				message,
				onTheFly );
		stopErrorProcessing();
	}

	/** Inform all the listeners */	
	public void notifyNoError( boolean onTheFly ) {
		if ( errors != null ) {
			for ( int i = 0; i < errors.size(); i++ ) {
				( ( ErrorListener )errors.get( i ) ).notifyNoError( onTheFly );				
			}
		}
		lastCaseInError = false;
	}

	/** Eliminate any trace of the last error. Should'nt be called */
	public void flushLastError() { 
		lastCaseInError = false;
		lastCaseOnTheFly = false;
		
	}

	/** @return <code>true</code> if an error exists */
	public boolean hasLastError() { return lastCaseInError; }
	/** @return <code>true</code> if the last error is due to an onthefly problem */
	public boolean hasLastErrorOnTheFly() { return lastCaseOnTheFly; }
	
	public void dispose() {
		errors = null;
	}
}
